@@ -17,12 +17,13 @@ other in minor ways, and should therefore be applied before others.
 
 Also note that not every augmentation makes sense to apply to every kind
 of payload (in particular, module import payloads don't make use of the
-"module" context slot, so augmentations like with_module_decorations
+"module" context slot, so augmentations like `with_module_decorations`
 can't be usefully applied to them).
 """
 
 import copy
 import sys
+import imp
 import io
 import re
 import traceback
@@ -61,7 +62,9 @@ AUGMENTATION_ORDER = [
     "with_timeout",
     "capturing_turtle_drawings",
     "capturing_wavesynth_audio",
+    "capturing_file_contents",
     "sampling_distribution_of_results",
+    "run_in_sandbox",
     "run_for_base_and_ref_values"
 ]
 """
@@ -130,8 +133,6 @@ def create_module_import_payload(
         file_path = context_utils.extract(context, "file_path")
         file_path = os.path.abspath(file_path)
         full_name = name_prefix + filename
-        tinfo = context_utils.extract(context, "task_info")
-        spec = tinfo["specification"]
 
         # Read the file
         with open(file_path, 'r', encoding="utf-8") as fin:
@@ -153,11 +154,13 @@ def create_module_import_payload(
             node = mast.parse(source, filename=full_name)
             errors = None
 
-        module = load.create_module_in_sandbox(
+        # Since this payload is already running inside a sandbox
+        # directory, we don't need to provide a sandbox argument here.
+        module = load.create_module_from_code(
             node,
             full_name,
-            sandbox_links=load.link_mapping(spec),
-            on_disk=file_path
+            on_disk=file_path,
+            sandbox=None
         )
 
         # Attempt to release any input capturing set up by optimism calls
@@ -237,7 +240,8 @@ def create_run_function_payload(
     If `copy_args` is set to True, deep copies of argument values will be
     made before they are passed to the target function (note that keyword
     argument keys are not copied, although they should be strings in any
-    case).
+    case). The "args" and "kwargs" slots will also get copies of the
+    arguments, not the original values.
 
     If the function name or any of the argument values (or keyword
     argument keys) are `potluck.context_utils.ContextualValue` instances,
@@ -268,7 +272,9 @@ def create_run_function_payload(
         fn = getattr(module, fname)
 
         real_posargs = []
+        initial_posargs = []
         real_kwargs = {}
+        initial_kwargs = {}
         for arg in posargs:
             if isinstance(arg, context_utils.ContextualValue):
                 try:
@@ -282,9 +288,12 @@ def create_run_function_payload(
                     raise
 
             if copy_args:
-                arg = copy.deepcopy(arg)
+                real_posargs.append(copy.deepcopy(arg))
+                initial_posargs.append(copy.deepcopy(arg))
 
-            real_posargs.append(arg)
+            else:
+                real_posargs.append(arg)
+                initial_posargs.append(arg)
 
         for key in kwargs:
             if isinstance(key, context_utils.ContextualValue):
@@ -311,15 +320,17 @@ def create_run_function_payload(
                     raise
 
             if copy_args:
-                value = copy.deepcopy(value)
-
-            real_kwargs[key] = value
+                real_kwargs[key] = copy.deepcopy(value)
+                initial_kwargs[key] = copy.deepcopy(value)
+            else:
+                real_kwargs[key] = value
+                initial_kwargs[key] = value
 
         return {
             "value": fn(*real_posargs, **real_kwargs),
             "function": fname,
-            "args": real_posargs,
-            "kwargs": real_kwargs
+            "args": initial_posargs,
+            "kwargs": initial_kwargs
         }
 
     return payload
@@ -673,6 +684,32 @@ def run_for_base_and_ref_values(
     return double_payload
 
 
+def run_in_sandbox(payload):
+    """
+    Returns a modified payload function which runs the provided base
+    payload, but first sets the current directory to the sandbox
+    directory specified by the provided context's "sandbox" slot.
+    Afterwards, it changes back to the original directory.
+
+    TODO: More stringent sandboxing?
+    """
+    def sandboxed_payload(context):
+        """
+        A payload function which runs a base payload within a specific
+        sandbox directory.
+        """
+        orig_cwd = os.getcwd()
+        try:
+            os.chdir(context_utils.extract(context, "sandbox"))
+            result = payload(context)
+        finally:
+            os.chdir(orig_cwd)
+
+        return result
+
+    return sandboxed_payload
+
+
 def with_setup(payload, setup):
     """
     Creates a modified payload which runs the given setup function
@@ -895,8 +932,7 @@ def create_mock_input(inputs, extra_policy="error"):
     from the given "inputs" sequence, and a reset function that resets
     the first function to the beginning of its inputs list.
 
-    The extra_inputs_policy specifies what happens if the inputs list
-    runs out:
+    The extra_policy specifies what happens if the inputs list runs out:
 
     - "loop" means that it will be repeated again, ad infinitum.
     - "hold" means that the last value will be returned for all
@@ -1579,6 +1615,94 @@ def capturing_turtle_drawings(payload, skip_reset=False, alt_text=None):
 # Wavesynth management #
 #----------------------#
 
+_PLAY_WAVESYNTH_TRACK = None
+"""
+The original wavesynth playTrack function, stored here temporarily while
+it's disabled via `disable_track_actions`.
+"""
+
+_SAVE_WAVESYNTH_TRACK = None
+"""
+The original wavesynth saveTrack function, stored here temporarily when
+saveTrack is disabled via `disable_track_actions`.
+"""
+
+
+def disable_track_actions():
+    """
+    Disables the `playTrack` and `saveTrack` `wavesynth` functions,
+    turning them into functions which accept the same arguments and
+    simply instantly return None. This helps ensure that students'
+    testing calls to `saveTrack` or `playTrack` don't eat up evaluation
+    time. Saves the original functions in the `_PLAY_WAVESYNTH_TRACK` and
+    `_SAVE_WAVESYNTH_TRACK` global variables.
+
+    Only saves original functions the first time it's called, so that
+    `reenable_track_actions` will work even if `disable_track_actions` is
+    called multiple times.
+
+    Note that you may want to use this function with
+    `specifications.add_module_prep` to ensure that submitted code
+    doesn't try to call `playTrack` or `saveTrack` during import and
+    waste evaluation time.
+    """
+    global _PLAY_WAVESYNTH_TRACK, _SAVE_WAVESYNTH_TRACK
+    import wavesynth
+    if _PLAY_WAVESYNTH_TRACK is None:
+        _PLAY_WAVESYNTH_TRACK = wavesynth.playTrack
+        _SAVE_WAVESYNTH_TRACK = wavesynth.saveTrack
+    wavesynth.playTrack = lambda wait=None: None
+    wavesynth.saveTrack = lambda filename: None
+
+
+def reenable_track_actions():
+    """
+    Restores the `saveTrack` and `playTrack` functions after
+    `disable_track_actions` has disabled them.
+    """
+    global _PLAY_WAVESYNTH_TRACK, _SAVE_WAVESYNTH_TRACK
+    import wavesynth
+    if _PLAY_WAVESYNTH_TRACK is not None:
+        wavesynth.playTrack = _PLAY_WAVESYNTH_TRACK
+        wavesynth.saveTrack = _SAVE_WAVESYNTH_TRACK
+        _PLAY_WAVESYNTH_TRACK = None
+        _SAVE_WAVESYNTH_TRACK = None
+
+
+def ensure_or_stub_simpleaudio():
+    """
+    Tries to import the `simpleaudio` module, and if that's not possible,
+    creates a stub module named "simpleaudio" which raises an attribute
+    error on any access attempt. The stub module will be inserted in
+    `sys.modules` as if it were `simpleaudio`.
+
+    Note that you may want to set this up as a prep function using
+    `specifications.add_module_prep` to avoid crashing if submitted code
+    tries to import `simpleaudio` (although it will still crash if
+    student code tries to use anything from `simpleaudio`).
+    """
+    # We also try to import simpleaudio, but set up a dummy module in its
+    # place if it's not available, since we don't need or want to play
+    # the sounds for grading purposes.
+    try:
+        import simpleaudio # noqa F401
+    except Exception:
+        def missing(name):
+            """
+            Fake getattr to raise a reasonable-seeming error if someone
+            tries to use our fake simpleaudio.
+            """
+            raise AttributeError(
+                "During grading, simpleaudio is not accessible. We have"
+                " disabled playTrack and saveTrack for testing purposes"
+                " anyway, and your code should not need to use"
+                " simpleaudio directly either."
+            )
+        fake_simpleaudio = imp.new_module("simpleaudio")
+        fake_simpleaudio.__getattr__ = missing
+        sys.modules["simpleaudio"] = fake_simpleaudio
+
+
 def capturing_wavesynth_audio(payload, just_capture=None, label=None):
     """
     Creates a modified version of the given payload which establishes
@@ -1604,6 +1728,11 @@ def capturing_wavesynth_audio(payload, just_capture=None, label=None):
     # available (we let any exception bubble out naturally).
     import wavesynth
 
+    # We do this here just in case student code attempts to use
+    # simpleaudio directly, since installing simpleaudio for evaluation
+    # purposes shouldn't be necessary.
+    ensure_or_stub_simpleaudio()
+
     def capturing_payload(context):
         """
         Resets all tracks state, runs a base payload, and then captures
@@ -1615,17 +1744,13 @@ def capturing_wavesynth_audio(payload, just_capture=None, label=None):
         wavesynth.resetTracks()
 
         # Disable playTrack and saveTrack
-        pT = wavesynth.playTrack
-        wavesynth.playTrack = lambda _=None: None
-        sT = wavesynth.saveTrack
-        wavesynth.saveTrack = lambda _: None
+        disable_track_actions()
 
         # Run the base payload
-        result = payload(context)
-
-        # Reenable playTrack and saveTrack
-        wavesynth.playTrack = pT
-        wavesynth.saveTrack = sT
+        try:
+            result = payload(context)
+        finally:
+            reenable_track_actions()
 
         # capture the descriptions of the notes in the current track
         if just_capture in (None, "notes"):
@@ -1653,3 +1778,206 @@ def capturing_wavesynth_audio(payload, just_capture=None, label=None):
         return result
 
     return capturing_payload
+
+
+#---------------------------------#
+# Miscellaneous harness functions #
+#---------------------------------#
+
+def report_argument_modifications(target, *args, **kwargs):
+    """
+    This function works as a test harness but doesn't capture the value
+    or output of the function being tested. Instead, it generates a text
+    report on whether or not each mutable argument to the function was
+    modified or not after the function is finished. It only checks
+    arguments which are are lists or dictionaries at the top level, so
+    its definition of modifiable is rather narrow.
+
+    (Note: the last two paragraphs of this docstring are picked up
+    automatically as rubric values for tests using this harness. fname
+    will be substituted in, which is why it appears in curly braces
+    below.)
+
+    Description:
+
+    <code>{fname}</code> must only modify arguments it is supposed to
+    modify.
+
+    We will call <code>{fname}</code> and check to make sure that the
+    values provided as arguments are not changed by the function, except
+    where such changes are explicitly required. Note that only mutable
+    values, like dictionaries or lists, may be modified by a function, so
+    this check is not applied to any string or number arguments.
+    """
+    # Identify mutable arguments
+    mposargs = [
+        i
+        for i in range(len(args))
+        if isinstance(args[i], (list, dict))
+    ]
+    mkwargs = [k for k in kwargs if isinstance(kwargs[k], (list, dict))]
+    if target.__kwdefaults__ is not None:
+        mkwdefaults = [k for k in target.__kwdefaults__ if k not in kwargs]
+    else:
+        mkwdefaults = []
+    nargs = target.__code__.co_argcount + target.__code__.co_kwonlyargcount
+    margnames = [target.__code__.co_varnames[:nargs][i] for i in mposargs]
+    mposnames = margnames[:len(mposargs)]
+    mposvals = [copy.deepcopy(args[i]) for i in mposargs]
+    mkwvals = [copy.deepcopy(kwargs[k]) for k in mkwargs]
+    mkwdefvals = {
+        k: copy.deepcopy(target.__kwdefaults__[k])
+        for k in mkwdefaults
+    }
+
+    # Call the target function
+    _ = target(*args, **kwargs)
+
+    # Report on which arguments were modified
+    result = ""
+
+    # Changes in positional argument values
+    for argindex, orig in zip(mposargs, mposvals):
+        final = args[argindex]
+        name = mposnames[argindex]
+        result += "Your code {} the value of the '{}' argument.\n".format(
+            "modified" if orig != final else "did not modify",
+            name
+        )
+
+    # Changes in keyword argument values
+    for name, orig in zip(mkwargs, mkwvals):
+        final = kwargs[name]
+        result += "Your code {} the value of the '{}' argument.\n".format(
+            "modified" if orig != final else "did not modify",
+            name
+        )
+
+    # Changes in values of unsupplied keyword arguments (i.e., changes to
+    # defaults, which if unintentional is usually bad!)
+    for name, orig in zip(mkwdefaults, mkwdefvals):
+        final = target.__kwdefaults__[name]
+        result += "Your code {} the value of the '{}' argument.\n".format(
+            "modified" if orig != final else "did not modify",
+            name
+        )
+
+    # The report by default will be compared against an equivalent report
+    # from the solution function, so that's how we figure out which
+    # arguments *should* be modified or not.
+    return result
+
+
+def returns_a_new_value(target, *args, **kwargs):
+    """
+    Checks whether or not the target function returns a value which is
+    new (i.e., not the same object as one of its arguments). Uses the
+    'is' operator to check for same-object identity, so it will catch
+    cases in which an object is modified and then returned. Returns a
+    string indicating whether or not a newly-constructed value is
+    returned.
+
+    Note: won't catch cases where the result is a structure which
+    *includes* one of the arguments. And does not check whether the
+    result is equivalent to one of the arguments, just whether it's
+    actually the same object or not.
+
+    (Note: the last two paragraphs of this docstring are picked up
+    automatically as rubric values for tests using this harness. fname
+    will be substituted in, which is why it appears in curly braces
+    below.)
+
+    Description:
+
+    <code>{fname}</code> must return a new value, rather than returning
+    one of its arguments.
+
+    We will call <code>{fname}</code> and check to make sure that the
+    value it returns is a new value, rather than one of the arguments it
+    was given (modified or not).
+    """
+    # Call the target function
+    fresult = target(*args, **kwargs)
+
+    # Check the result against each of the arguments
+    nargs = target.__code__.co_argcount + target.__code__.co_kwonlyargcount
+    for argindex, argname in enumerate(target.__code__.co_varnames[:nargs]):
+        if argindex < len(args):
+            # a positional argument
+            argval = args[argindex]
+        else:
+            # a keyword argument (possibly defaulted via omission)
+            argval = kwargs.get(argname, target.__kwdefaults__[argname])
+
+        if fresult is argval:
+            return (
+                "Returned the '{}' argument (possibly with modifications)."
+            ).format(argname)
+
+    # Since we didn't return in the loop above, there's no match
+    return "Returned a new value."
+
+
+#------------------#
+# File I/O Helpers #
+#------------------#
+
+def file_contents_setter(filename, contents):
+    """
+    Returns a setup function (use with `with_setup`) which replaces the
+    contents of the given file with the given contents. Be careful,
+    because this will happily overwrite any file. If the desired contents
+    is a bytes object, the file will be written in binary mode to contain
+    exactly those bytes, otherwise contents should be a string.
+    """
+    def setup_file_contents(context):
+        """
+        Returns the provided context as-is, but before doing so, writes
+        data to a specific file to set it up for the coming test.
+        """
+        if isinstance(contents, bytes):
+            with open(filename, 'wb') as fout:
+                fout.write(contents)
+        else:
+            with open(filename, 'w') as fout:
+                fout.write(contents)
+        return context
+
+    return setup_file_contents
+
+
+def capturing_file_contents(payload, filename, binary=False):
+    """
+    Captures the entire contents of the given filename as a string (or a
+    bytes object if binary is set to True), and stores it in the
+    "output_file_contents" context slot. Also stores the file name of the
+    file that was read in in the "output_filename" slot.
+    """
+    def capturing_payload(context):
+        """
+        Runs a base payload and then reads the contents of a specific
+        file, adding that data as a "output_file_contents" context slot
+        and also adding an "output_filename" slot holding the filename
+        that was read from.
+        """
+        # Run base payload
+        result = payload(context)
+
+        # Record filename in result
+        result["output_filename"] = filename
+
+        # Decide on open flags
+        if binary:
+            flags = 'rb'
+        else:
+            flags = 'r'
+
+        with open(filename, flags) as fin:
+            file_contents = fin.read()
+
+        # Add file contents
+        result["output_file_contents"] = file_contents
+
+        return result
+
+    return capturing_payload
