@@ -7,6 +7,7 @@ render.py
 import os
 import sys
 import json
+import base64
 
 import markdown
 import jinja2
@@ -190,10 +191,13 @@ def render_markdown(md_source):
 
     try:
         # Try to render with extensions
-        return markdown.markdown(md_source, extensions=[x])
+        result = markdown.markdown(md_source, extensions=[x])
+        return result
     except Exception:
         # Backup: try without extensions, since in some cases they don't
         # work I guess :(
+        print("Error rendering markdown with extensions:")
+        print(html_tools.string_traceback())
         return markdown.markdown(md_source, extensions=[])
 
 
@@ -352,17 +356,7 @@ def render_instructions(
           + len(other_unlinked)
         )
 
-        if len(invalid_links) == 0:
-            print("All links to rubric goals were valid.")
-        else:
-            print(
-                (
-                    "{} link(s) to rubric goal(s) were invalid:"
-                ).format(len(invalid_links))
-            )
-            for inv in sorted(invalid_links):
-                print('  ' + inv)
-
+        # Report on goals not linked
         if total_unlinked == 0:
             print("All goals were linked to from the instructions.")
         else:
@@ -387,6 +381,18 @@ def render_instructions(
                 for goal in sorted(other_unlinked):
                     print('  ' + goal)
 
+        # Report on invalid links last
+        if len(invalid_links) == 0:
+            print("All links to rubric goals were valid.")
+        else:
+            print(
+                (
+                    "{} link(s) to rubric goal(s) were invalid:"
+                ).format(len(invalid_links))
+            )
+            for inv in sorted(invalid_links):
+                print('  ' + inv)
+
     # Save to the output file
     with open(output_file, 'w', encoding="utf-8") as fout:
         fout.write(rendered)
@@ -408,18 +414,41 @@ def augment_report(report, blank=False):
     - 'js' - JavaScript code for reports
 
     For each file dictionary in the 'files' slot, it also adds:
-        'code_html' - a rendering of the file's submitted code as HTML
-    (but only if there's a "code" slot).
+
+    - 'code_html' - a rendering of the file's submitted code as HTML
+
+    (but only if that file has a "code" slot).
     """
     if "files" in report:
         for file_entry in report["files"]:
-            if "code" in file_entry:
+            if "original_code" in file_entry:
+                # We render the fixed code, but encode the original
+                file_entry["base64"] = base64.standard_b64encode(
+                    file_entry["original_code"].encode("utf-8")
+                ).decode("utf-8")
                 file_entry["code_html"] = render_code(
                     report["taskid"],
                     file_entry["filename"],
                     file_entry["code"]
                 )
-            # TODO: Something else for "raw" files?
+            elif "code" in file_entry:
+                # Render & encode the same code
+                file_entry["base64"] = base64.standard_b64encode(
+                    file_entry["code"].encode("utf-8")
+                ).decode("utf-8")
+                file_entry["code_html"] = render_code(
+                    report["taskid"],
+                    file_entry["filename"],
+                    file_entry["code"]
+                )
+            elif "raw" in file_entry:
+                # Encode raw data but don't render anything
+                # We won't try to present the file data in any fancy way
+                file_entry["base64"] = base64.standard_b64encode(
+                    file_entry["raw"].encode("utf-8")
+                ).decode("utf-8")
+            # File entires should all have either 'raw' or 'code'
+            # slots, but if one doesn't, we'll just leave it alone
 
     if "table" in report:
         report['rendered_table'] = render_rubric_table(
