@@ -75,11 +75,8 @@ module (such as `provide_inputs` as shown above) to control exactly how
 the code is run.
 """
 
-import pprint
 import textwrap
 import re
-import io
-import base64
 import os
 
 import pygments
@@ -105,9 +102,176 @@ Line width we'll attempt to hit for formatting output nicely.
 """
 
 
-def pprint_with_prefix(obj, prefix, indentation_level=0):
+def wrap_str(string, width):
     """
-    Returns a string representing the given object, as pprint.pformat
+    Returns a list of strings, which when combined form the given
+    string, where none is longer than the given width. Lines are broken
+    at spaces whenever possible, but may be broken in the middle of
+    words when a single word is longer than the target width. This
+    function uses a greedy algorithm that doesn't optimize the evenness
+    of the right margin. \\n and \\r are represented by those character
+    pairs, but other characters are as-is.
+    """
+    escaped = string.replace('\n', '\\n')
+    escaped = escaped.replace('\r', '\\r')
+    words = escaped.split(' ')
+    result = []
+    while len(words) > 0: # until we've used up the words
+        # Figure out how many words would fit on the next line
+        i = 0
+        # + i accounts for spaces we'll need
+        while len(words[:i + 1]) + (i + 1) < width:
+            i += 1
+
+        # Back off to the number that fits
+        i -= 1
+
+        # first word doesn't fit!
+        if i == -1:
+            first = words[0]
+            # Chop up the first word and put the pieces into the result
+            pieces = len(first) // width
+            for j in range(pieces):
+                result.append(first[width * j:width * (j + 1)])
+
+            # Handle leftovers
+            rest = first[width * pieces:]
+            # Note that even empty leftovers serve as a placeholder to
+            # ensure that a space gets added.
+            words[0] = rest # put rest back into words, replacing old
+        else:
+            # at least one word fits
+            line = ' '.join(words[:i + 1]) + ' '
+            result.append(line)
+            words = words[i + 1:]
+
+    # Remove the extra space introduced after the end of the last word
+    # Works even for an empty string input
+    result[-1] = result[-1][:-1]
+
+    return result
+
+
+def wrapped_repr(obj, width=FMT_WIDTH, indentation=1, prefix='', memo=None):
+    """
+    Returns a string representing the given object, whose lines are not
+    longer than the given width, and which uses the given number of
+    spaces to indent sub-objects when necessary. Replacement for
+    pprint.pformat, which does NOT reorder dictionary keys. May violate
+    the width restriction when given a complex object that it doesn't
+    know how to wrap whose repr alone is wider than the specified width.
+
+    A prefix may be supplied, which will be added to the first line of
+    the given representation as-is. The prefix itself will not be subject
+    to wrapping, but the first line of the result will in some cases wrap
+    earlier to compensate for the prefix. The prefix should not include
+    any newlines or carriage returns.
+
+    The memo helps avoid problems with recursion, and doesn't need to be
+    provided explicitly.
+    """
+    # Create a memo if we need to
+    if memo is None:
+        memo = set()
+
+    # Check the memo
+    if id(obj) in memo:
+        return '...'
+
+    # Add ourselves to the memo
+    memo.add(id(obj))
+
+    simple = prefix + repr(obj)
+
+    if len(simple) <= width: # no need to do anything fancy
+        return simple
+
+    indent = ' ' * indentation
+    between = '\n' + indent
+
+    broken = repr(obj)
+    if len(broken) < width - indentation:
+        return prefix + '\\' + between + broken
+
+    elif type(obj) == str: # need to wrap this string
+        lines = wrap_str(obj, width - (indentation + 2))
+        return (
+            prefix + '(' + between
+          + between.join('"' + line + '"' for line in lines)
+          + '\n)'
+        )
+
+    elif type(obj) == bytes: # need to wrap these bytes
+        lines = wrap_str(obj, width - (indentation + 3))
+        return (
+            prefix + '(' + between
+          + between.join('b"' + line + '"' for line in lines)
+          + '\n)'
+        )
+
+    elif type(obj) in (list, tuple, set): # need to wrap each item
+        if type(obj) in (list, tuple):
+            ldelim, rdelim = str(type(obj)())
+        elif len(obj) == 0:
+            return prefix + 'set()' # too bad if width is < 5
+        else:
+            ldelim = '{'
+            rdelim = '}'
+
+        result = prefix + ldelim + between
+        for item in obj:
+            result += textwrap.indent(
+                wrapped_repr(item, width - indentation, indentation),
+                indent
+            ) + between
+
+        return result[-len(between):] + '\n' + rdelim
+
+    elif type(obj) == dict: # need to wrap keys and values
+        indent = ' ' * indentation
+        between = '\n' + indent
+
+        result = prefix + '{' + between
+        for key in obj:
+            key_repr = wrapped_repr(
+                key,
+                width - (indentation + 1),
+                indentation
+            )
+            val = obj[key]
+            val_repr = wrapped_repr(
+                val,
+                width - indentation * 2,
+                indentation
+            )
+
+            # Can we fit on one line?
+            if len(indent) + len(key_repr) + len(val_repr) + 3 <= width:
+                result += key_repr + ': ' + val_repr + ',' + between
+            elif val_repr[1] == '\n': # ldelim/newline start
+                result += (
+                    key_repr + ':' + val_repr[0] + '\n'
+                  + textwrap.indent(val_repr[2:], indent)
+                  + ',' + between
+                )
+            else:
+                result += key_repr + ':\n' + textwrap.indent(
+                    val_repr,
+                    indent * 2
+                ) + ',' + between
+
+        # Chop off final between and add closing brace after a newline
+        result = result[:-len(between)] + '\n}'
+
+        return result
+
+    else: # Not sure how to wrap this, so we give up...
+        return prefix + repr(obj)
+
+
+def wrapped_with_prefix(obj, prefix, indentation_level=0):
+    """
+    Returns a string representing the given object, as `wrapped_repr`
     would, except that:
 
     1. The first line includes the given prefix (as a string without
@@ -116,24 +280,18 @@ def pprint_with_prefix(obj, prefix, indentation_level=0):
     2. The entire representation is indented by the given indentation
        level (one space per level).
     """
-    result = pprint.pformat(
-        (prefix, obj),
-        indent=1,
-        width=FMT_WIDTH - indentation_level
+    rep = wrapped_repr(
+        obj,
+        width=FMT_WIDTH - indentation_level,
+        indentation=1,
+        prefix=prefix
     )
 
-    # Strip off placeholder at beginning
-    begins = len("({},".format(pprint.pformat(prefix)))
-    rep = result[begins:-1] # one extra paren at end
-
     # remove extra indentation from tuple
     rep = textwrap.dedent(rep)
 
-    # remove possible extra newline at start
-    rep = rep.strip()
-
     # restore indent and add prefix
-    return textwrap.indent(prefix + rep, ' ' * indentation_level)
+    return textwrap.indent(rep, ' ' * indentation_level)
 
 
 #--------------#
@@ -242,7 +400,7 @@ class Snippet(specifications.TestGroup):
         caption,
         tests,
         postscript='',
-        glammers=None,
+        glamers=None,
         ignore_errors=False
     ):
         """
@@ -253,7 +411,7 @@ class Snippet(specifications.TestGroup):
         An optional postscript is markdown source to be placed after the
         code block.
 
-        An optional list of glammer functions may be provided. These will
+        An optional list of glamer functions may be provided. These will
         be given individual pieces of HTML code and the associated
         context those were rendered from and their result value will
         replace the HTML code to be used; they are applied in-order.
@@ -270,7 +428,7 @@ class Snippet(specifications.TestGroup):
         self.title = title
         self.caption = caption
         self.postscript = postscript
-        self.glammers = glammers or []
+        self.glamers = glamers or []
         self.ignore_errors = ignore_errors
 
         # Final edit function to manipulate the context to be rendered
@@ -478,8 +636,8 @@ class Snippet(specifications.TestGroup):
               + context.get("error_log", '<missing>')
             )
         result = self.render_context(context)
-        for glammer in self.glammers:
-            result = glammer(result, context)
+        for glamer in self.glamers:
+            result = glamer(result, context)
 
         return result
 
@@ -514,7 +672,7 @@ class Snippet(specifications.TestGroup):
 
     def show_image(self):
         """
-        Adds a glammer that augments output by appending a "Image "
+        Adds a glamer that augments output by appending a "Image "
         prompt followed by an image element which shows the contents of
         the "image" context slot. To establish an "image" slot, call a
         method like
@@ -522,19 +680,33 @@ class Snippet(specifications.TestGroup):
 
         Returns self for chaining.
         """
-        self.glammers.append(append_image_result)
+        self.glamers.append(append_image_result)
         return self
 
     def play_audio(self):
         """
-        Adds a glammer that augments output by appending an "Audio "
+        Adds a glamer that augments output by appending an "Audio "
         prompt followed by an audio element which plays the contents of
         the "audio" context slot. To make an "audio" slot, call a method
         like `potluck.specifications.HasPayload.capture_wavesynth`.
 
         Returns self for chaining.
         """
-        self.glammers.append(append_audio_result)
+        self.glamers.append(append_audio_result)
+        return self
+
+    def show_file_contents(self):
+        """
+        Adds a glamer that augments output by appending a "File"
+        prompt followed by a text box showing the filename and the
+        contents of that file. Requires 'output_filename' and
+        'output_file_contents' slots, which could be established using a
+        method like
+        `potluck.specifications.HasPayload.capture_file_contents`
+
+        Returns self for chaining.
+        """
+        self.glamers.append(append_file_contents)
         return self
 
 
@@ -595,9 +767,9 @@ class Variables(Snippet):
         varname = context["variable"]
         value = context["value"]
 
-        # format value's repr using pprint, but leaving room for varname
-        # = at beginning of first line
-        rep = pprint_with_prefix(value, varname + ' = ')
+        # format value's repr using wrapped_repr, but leaving room for
+        # varname = at beginning of first line
+        rep = wrapped_with_prefix(value, varname + ' = ')
 
         # Use pygments to generate HTML markup for our assignment
         markup = highlight_code(rep)
@@ -697,6 +869,9 @@ class FunctionCalls(Snippet):
         calling a function. Includes a Jupyter-style prompt for input as
         well as the return value.
         """
+        if "error" in context:
+            print("Error during context creation:")
+            print(context["error"])
         fname = context["function"]
         value = context["value"]
         args = context["args"]
@@ -707,10 +882,10 @@ class FunctionCalls(Snippet):
         # Figure out representations of each argument
         argreps = []
         for arg in args:
-            argreps.append(pprint_with_prefix(arg, '', 1))
+            argreps.append(wrapped_with_prefix(arg, '', 1))
 
         for kw in kwargs:
-            argreps.append(pprint_with_prefix(kwargs[kw], kw + "=", 1))
+            argreps.append(wrapped_with_prefix(kwargs[kw], kw + "=", 1))
 
         # Figure out full function call representation
         oneline = "{}({})".format(
@@ -751,7 +926,7 @@ class FunctionCalls(Snippet):
 
         # Highlight the return value
         if value is not None:
-            value = highlight_code(pprint_with_prefix(value, ""))
+            value = highlight_code(wrapped_with_prefix(value, ""))
 
             result += (
                 '<span class="prompt output">Out[]:</span>'
@@ -856,7 +1031,7 @@ class Blocks(Snippet):
 
         # Highlight the return value
         if value is not None:
-            value = highlight_code(pprint_with_prefix(value, ""))
+            value = highlight_code(wrapped_with_prefix(value, ""))
 
             result += (
                 '<span class="prompt output">Out[]:</span>'
@@ -869,7 +1044,7 @@ class Blocks(Snippet):
 class Fakes(Snippet):
     """
     One or more fake snippets, which format code, printed output, stderr
-    output, and a result value (and possibly glammer additions) but each
+    output, and a result value (and possibly glamer additions) but each
     of these things is simply specified ahead of time. Note that you want
     to avoid using this if at all possible, because it re-creates the
     problems that this module was trying to solve (namely, examples
@@ -883,7 +1058,7 @@ class Fakes(Snippet):
         of fake context dictionary objects to be rendered as if they came
         from a real test. Each dictionary must contain a "code" slot, and
         may contain "value", "output", and/or "error_log" keys. If using
-        glammers, relevant keys should be added directly to these fake
+        glamers, relevant keys should be added directly to these fake
         contexts.
 
         If "value" or "output" slots are missing, defaults of None and ''
@@ -971,7 +1146,7 @@ class Fakes(Snippet):
 
         # Highlight the return value
         if value is not None:
-            value = highlight_code(pprint_with_prefix(value, ""))
+            value = highlight_code(wrapped_with_prefix(value, ""))
 
             result += (
                 '<span class="prompt output">Out[]:</span>'
@@ -981,9 +1156,9 @@ class Fakes(Snippet):
         return result
 
 
-#----------#
-# Glammers #
-#----------#
+#---------#
+# Glamers #
+#---------#
 
 def append_image_result(markup, context):
     """
@@ -1008,12 +1183,7 @@ def append_image_result(markup, context):
         context.get("output", "no alt text available")
     )
 
-    src_bytes = io.BytesIO()
-    image.save(src_bytes, format="PNG")
-    data = base64.b64encode(src_bytes.getvalue()).decode("utf-8")
-    img_tag = (
-        '<img class="example" src="data:image/png;base64,{}" alt="{}">'
-    ).format(data, alt)
+    img_tag = html_tools.html_image(image, alt, ["example"])
     return (
         markup
       + '<span class="prompt special">Image </span>'
@@ -1029,7 +1199,8 @@ def append_audio_result(markup, context):
 
     Note that this can result in a pretty large string if the WAV format
     is used, since the string needs to be base64-encoded and WAV is
-    uncompressed :(
+    uncompressed (and we double the size by including the data URL twice)
+    :(
 
     TODO: Maybe use Ogg/Vorbis?
 
@@ -1042,7 +1213,7 @@ def append_audio_result(markup, context):
     "label" slot which will be used invisibly as an aria-label property
     of the audio element; if absent no aria-label will be attached.
     """
-    # Short-circuit unless we've got an image
+    # Short-circuit unless we've got audio
     if "audio" not in context:
         return markup
 
@@ -1051,20 +1222,7 @@ def append_audio_result(markup, context):
     data = audio["data"]
     label = audio.get("label")
 
-    data = base64.b64encode(data).decode("utf-8")
-    data_url = 'data:{};base64,{}'.format(mime, data)
-    audio_tag = '''\
-<audio controls{label}>{source}
-(Your browser does not support the audio element. Use this link to download the file and play it using a media player: <a href="{data_url}">download audio</a>)
-</audio>'''.format( # noqa E501
-        label=(
-            ' aria-label="{}"'.format(label)
-            if label is not None
-            else ''
-        ),
-        source='<source type="{}" src="{}">'.format(mime, data_url),
-        data_url=data_url
-    )
+    audio_tag = html_tools.html_audio(data, mime, label)
     return (
         markup
       + '\n<span class="prompt special">Audio </span>'
@@ -1072,6 +1230,38 @@ def append_audio_result(markup, context):
     )
 
 
+def append_file_contents(markup, context):
+    """
+    Given some HTML markup and a context dictionary, turns the
+    "output_filename" and "output_file_contents" context slots into HTML
+    markup displaying the contents of that file, which gets appended to
+    the given markup and returned.
+
+    Note that this is intended to display the contents of text files.
+
+    If either of the "output_filename" or "output_file_contents" context
+    slots are missing, the given markup is returned unmodified.
+    """
+    # Short-circuit unless we've got output file contents
+    if (
+        "output_filename" not in context
+     or "output_file_contents" not in context
+    ):
+        return markup
+
+    filename = context["output_filename"]
+    contents = context["output_file_contents"]
+
+    return (
+        markup
+      + (
+            '\n<span class="prompt special">File</span>'
+        )
+      + '<div class="snippet-filename">{}</div>\n'.format(filename)
+      + '<div class="snippet-file-contents">{}</div>\n'.format(contents)
+    )
+
+
 #--------#
 # Lookup #
 #--------#
