@@ -51,6 +51,7 @@ RUBRIC_EXPECTS = {
         "exhibits the correct behavior"
     ],
     "filesTest": [ "<title>filesTest Rubric</title>" ],
+    "varsTest": [ "<title>varsTest Rubric</title>" ],
 }
 
 # Expectations about reports
@@ -81,8 +82,11 @@ REPORT_EXPECTS = {
     },
     "filesTest": {
         "perfect": { "evaluation": "excellent" },
-        #"imperfect": { "evaluation": "incomplete" },
-        # TODO: This!
+        "imperfect": { "evaluation": "incomplete" },
+    },
+    "varsTest": {
+        "perfect": { "evaluation": "excellent" },
+        "imperfect": { "evaluation": "partially complete" },
     }
 }
 
@@ -97,7 +101,8 @@ REPORT_EXPECTS = {
         "sceneTest",
         "freedomTest",
         "snippetsTest",
-        "filesTest"
+        "filesTest",
+        "varsTest",
     ]
 )
 def taskid(request):
